/*******************************************************************************
                     Computación Paralela 2012 - FaMAF
                      Universidad Nacional de Córdoba

Ejercicio 1 - La traspuesta contraataca
=======================================

a) El kernel para trasponer que implementamos en el práctico pasado tiene malos
   patrones de acceso a memoria por warp, que hacen que sea lento. Mejore la
   performance del kernel haciendo que cada bloque de valores se lea y escriba
   secuencialmente por warp, y la trasposición se realice en shared memory.
b) Investigue en la CUDA C Programmer's Guide cómo funcionan los bancos de
   shared memory, y calcule cómo un warp ejecutando su kernel accede a estos
   recursos. Diseñe e implemente una variante del kernel que mitigue este
   problema.
c) Compare los contadores l1_shared_bank_conflict de ambos kernels usando la
   interfaz de línea de comando o nvvp. Se solucionó del todo el problema?
d) Usar un tamaño de bloque más grande reduce la cantidad de conflictos, pero
   se pierde ocultamiento de latencia. Aplique thread coarsening e implemente
   una versión del kernel donde cada thread procesa más de un elemento para poder
   usar bloques de threads pequeños que procesan una gran cantidad de elementos.
e) Cuál es la mayor cantidad de elementos que podemos procesar en un bloque de
   256 threads? Utilice el CUDA Occupancy Calculator para ver cómo los distintos
   recursos utilizados por un kernel limitan la ocupación del SM.


********************************************************************************/

#include <cuda.h>
#include <cutil.h>
#include <stdio.h>
#include <stdlib.h>

#include "colorramp.h"
#include "sdlstuff.h"

#define N 523
#define SIZE (N * N)

#define BLOCK_X 16
#define BLOCK_Y 16


// pinta la imagen con una gradiente donde se ve bien la traspuesta
__global__ void gradient_2d(rgba32 * image, size_t n) {
    uint x = blockIdx.x * blockDim.x + threadIdx.x;
    uint y = blockIdx.y * blockDim.y + threadIdx.y;
    uint idx = y * n + x;
    if (x < n && y < n) {
        float u = (float) x / n;
        float v = (float) y / n;
        image[idx] = color_2d(u, v);
    }
}


// Transpone una matriz cuadrada n x n
__global__ void transpose_2d(const rgba32 * src, size_t n, rgba32 * dst) {
    uint x = blockIdx.x * blockDim.x + threadIdx.x;
    uint y = blockIdx.y * blockDim.y + threadIdx.y;
    if (x < n && y < n) {
        size_t src_idx = x + y * n;
        size_t dst_idx = y + x * n;
        dst[dst_idx] = src[src_idx];
    }
}


int main(int argc, char ** argv) {

    rgba32 * device_orig;
    rgba32 * device_trans;
    CUDA_SAFE_CALL(cudaMalloc(&device_orig, SIZE * sizeof(rgba32)));
    CUDA_SAFE_CALL(cudaMalloc(&device_trans, SIZE * sizeof(rgba32)));

    // pintamos las imagen de negro para ver errores
    CUDA_SAFE_CALL(cudaMemset(device_trans, 0, SIZE * sizeof(rgba32)));

    // dimensiones de grid 2d
    dim3 block_2d(BLOCK_X, BLOCK_Y);
    dim3 grid_2d((N + block_2d.x - 1) / block_2d.x, (N + block_2d.y - 1) / block_2d.y);

    // eventos para tomar el tiempo
    cudaEvent_t start, stop;
    CUDA_SAFE_CALL(cudaEventCreate(&start));
    CUDA_SAFE_CALL(cudaEventCreate(&stop));

    // encolar coloreo de la imagen
    gradient_2d<<<grid_2d, block_2d>>>(device_orig, N);

    // encolar evento para el inicio de transpose_2d
    CUDA_SAFE_CALL(cudaEventRecord(start));

    // encolar kernel
    transpose_2d<<<grid_2d, block_2d>>>(device_orig, N, device_trans);

    // encolar evento de fin de tranpose_2d
    CUDA_SAFE_CALL(cudaEventRecord(stop));

    // esperar que se termine de procesar stop al final de la cola
    CUDA_SAFE_CALL(cudaEventSynchronize(stop));

    CUT_CHECK_ERROR("Falló algo: ");

    // tomar el tiempo
    float elapsed;
    CUDA_SAFE_CALL(cudaEventElapsedTime(&elapsed, start, stop));
    printf("transpose_2d: %f ms\n", elapsed);

    // pedir memoria del host para la imagen
    rgba32 * host_orig = (rgba32 *) malloc(SIZE * sizeof(rgba32));
    rgba32 * host_trans = (rgba32 *) malloc(SIZE * sizeof(rgba32));
    CUDA_SAFE_CALL(cudaMemcpy(host_orig, device_orig, SIZE * sizeof(rgba32), cudaMemcpyDefault));
    CUDA_SAFE_CALL(cudaMemcpy(host_trans, device_trans, SIZE * sizeof(rgba32), cudaMemcpyDefault));

#ifndef HEADLESS
    // inicializar sdl
    sdls_init(2 * N + 1, N);

    // dibujar en pantalla
    sdls_drawrectangle(0, 0, N, N, host_orig);
    sdls_drawrectangle(N + 1, 0, N, N, host_trans);

    printf("<ENTER> para salir\n");
    getchar();
#else
    for (uint y = 0; y < N; ++y) {
        for (uint x = 0; x < N; ++x) {
            if (!rgba32_isequal(&host_orig[y * N + x], &host_trans[x * N + y])) {
                printf("x:%u y:%u mal transpuesta\n", x, y);
            }
        }
    }
#endif

    // limpieza
    free(host_orig);
    free(host_trans);
    CUDA_SAFE_CALL(cudaFree(device_orig));
    CUDA_SAFE_CALL(cudaFree(device_trans));
    CUDA_SAFE_CALL(cudaEventDestroy(start));
    CUDA_SAFE_CALL(cudaEventDestroy(stop));

    return 0;
}
